# This file is part of PyGeBR library.
#
# PyGeBR is a library to create, merge, edit, inspect and run
# processing flows. It was originally designed for seismic
# data processing, but it extends to other data as well.
#
# PyGeBR is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyGeBR is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyGeBR.  If not, see <https://www.gnu.org/licenses/>
#
# Copyright 2021-2022 Ricardo Biloti <biloti@unicamp.br>
#                     Eduardo Filpo <efilpo@gmail.com>
#

import json as JSON
import ipywidgets as W

#-------------------------------------------------------------------------------
class Person:
    """Class used to represent someone to credit for.

Any program, flow or menu in PyGeBR may have authors, which
are represented as a list of Person.

Attributes
----------

json: dict
    Dictionary representing a person

widget: dict
    Dictionary holding an widget to edit person's attributes

name: str
    Person's name

email: str
    Person's email

institution: str
    Person's affiliation

homepage: str
    Person's or institution's website
"""

    def __init__ (self, json=None, name=None, email=None,
        institution=None, homepage=None):
        """
        Parameters
        ----------

        json: dictionary (optional)
            A dictionary to represent a person.

        name: str (optional)
            Person's name

        email: str (optional)
            Person's email

        institution: str (optional)
            Person's affiliation

        homepage: str (optional)
            Person's or institution's website
        """

        if json is None:
            self.json = {}
        else:
            if not isinstance(json, dict):
                raise Exception("Argument should be a dictionary")
            self.json = json

        self.widget = {}

        if not name is None:
            self.name(name)
        if not email is None:
            self.email(email)
        if not institution is None:
            self.institution(institution)
        if not homepage is None:
            self.homepage(homepage)

    def has_name (self):
        """
        Return whether a pearson has name set
        """

        return 'name' in self.json

    def name(self, value=None):
        """
        Set or return person's name

        Parameter
        ---------

        value: str (optional)
            value to set as person's name
        """
        if value is None:
            return self.json['name'] if self.has_name() else ""

        if not isinstance(value, str):
            raise Exception("Argument should be a string")

        if (not value) and 'name' in self.json:
            del self.json['name']
        else:
            self.json['name'] = value

    def has_email(self):
        """
        Return whether a pearson has email set
        """
        return 'email' in self.json

    def email(self, value=None):
        """
        Set or return person's email

        Parameter
        ---------

        value: str (optional)
            value to set as person's email
        """
        if value is None:
            return self.json['email'] if self.has_email() else ""

        if not isinstance(value, str):
            raise Exception("Argument should be a string")

        if (not value) and 'email' in self.json:
            del self.json['email']
        else:
            self.json['email'] = value

    def has_institution (self):
        """
        Return whether a pearson has institution set
        """
        return 'institution' in self.json

    def institution(self, value=None):
        """
        Set or return person's affiliation

        Parameter
        ---------

        value: str (optional)
            value to set as person's affiliation
        """
        if value is None:
            return self.json['institution'] if self.has_institution() else ""

        if not isinstance(value, str):
            raise Exception("Argument should be a string")

        if (not value) and 'institution' in self.json:
            del self.json['institution']
        else:
            self.json['institution'] = value

    def has_homepage (self):
        """
        Return whether a pearson has homepage set
        """
        return 'homepage' in self.json

    def homepage(self, value=None):
        """
        Set or return person's or institutions' website

        Parameter
        ---------

        value: str (optional)
            value to set as website
        """
        if value is None:
            return self.json['homepage'] if self.has_homepage() else ""

        if not isinstance(value, str):
            raise Exception("Argument should be a string")

        if (not value) and 'homepage' in self.json:
            del self.json['homepage']
        else:
            self.json['homepage'] = value

    def __str__(self):
        """
        Returns a multiline string, descring the person.
        """
        line = self.name()

        if self.has_email():
            if line:
                line = line + " "

            line = line + "<" + self.json['email'] +">"

        if line:
            line = line + "\n"

        if self.has_institution():
            line = line + self.institution() + "\n"

        if self.has_homepage():
            line = line + self.homepage() + "\n"

        return line[0:-1]

    def widget_construct(self):
        """
        Construct widget to edit person's attributes
        """

        if self.widget != {}:
            return

        self.widget['name'] = W.Text(placeholder="Name", value=self.name())
        self.widget['email'] = W.Text(placeholder="Email address", value=self.email())
        self.widget['institution'] = W.Text(placeholder="Institution", value=self.institution())
        self.widget['homepage'] = W.Text(placeholder="Homepage",value=self.homepage())

        self.widget['line'] = W.VBox([W.HBox([self.widget['name'],self.widget['email']]),
                         W.HBox([self.widget['institution'],self.widget['homepage']])],
                         layout=W.Layout(margin="0px 0px 10px 0px"))

    def w_get(self):
        """
        Set person's attributes from widget entries' value
        """

        self.name(self.widget['name'].value)
        self.email(self.widget['email'].value)
        self.institution(self.widget['institution'].value)
        self.homepage(self.widget['homepage'].value)

    def w_set(self):
        """
        Set widget entries' value from person's attributes
        """

        self.widget['name'].value = self.name()
        self.widget['email'].value = self.email()
        self.widget['institution'].value = self.institution()
        self.widget['homepage'].value = self.homepage()

    def w_line(self):
        """
        Return a widget to edit the person's attributes.
        The widget is meant to be included into a VBox.
        """
        if self.widget == {}:
            self.widget_construct()

        return self.widget['line']



# This file is part of PyGeBR library.
#
# PyGeBR is a library to create, merge, edit, inspect and run
# processing flows. It was originally designed for seismic
# data processing, but it extends to other data as well.
#
# PyGeBR is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyGeBR is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyGeBR.  If not, see <https://www.gnu.org/licenses/>
#
# Copyright 2021-2022 Ricardo Biloti <biloti@unicamp.br>
#                     Eduardo Filpo <efilpo@gmail.com>
#

import os
import json as JSON
from .person import Person

class Setup:
    """A class used to setup basic configuration parameters for PyGeBR

An instance for this class can setup or query basic parameters
for PyGeBR, like the default author for new flows made from
menus as well as search path for menus.
"""

    def __init__ (self, authors = None, menudirs = None):
        """
        Loads a setup saved on $HOME/.config/pygebr/config.json
        or instantiate a new one.

        Parameters
        ---------
        authors: Person or list of Person (optional)
            When provided set the list of author.

        menudirs: str or list of str (optional)
            When provided set the list of search paths.
        """

        self.configDir = os.path.join(os.environ['HOME'],'.config/pygebr')
        if not os.path.isdir(self.configDir):
            os.mkdir(self.configDir)

        self.configFile = os.path.join(self.configDir,'config.json')
        if os.path.isfile(self.configFile):
            with open(self.configFile) as fp:
                self.config = JSON.loads(fp.read())
        else:
            self.config = {"authors": [], "menudirs": []}

        if authors is not None:
            self.authors(authors)

        if menudirs is not None:
            self.menudirs(menudirs)

    def __str__(self):
        """
        Print the setup.
        """

        output = ""
        for person in self.authors():
            output = output + str(person) + "\n"

        output = output + "\nLook up for menus in:\n"
        for path in self.menudirs():
            output = output + path + "\n"

        return output

    def authors(self, authors=None):
        """
        Set or returns default authors for new flows.

        Parameter
        ---------
        authors: Person or list of Person
        When provided, set default authors.
        """

        if authors is None:
            alist = []
            for p in self.config['authors']:
                alist.append(Person(json=p))
            return alist

        if isinstance(authors, list):
            people = []
            for person in authors:
                if isinstance(person, Person):
                    people.append(person.json)
                else:
                    raise TypeError("authors is not of type list of Person")

            self.config['authors'] = people
        else:
            if isinstance(authors, Person):
                self.config['authors'] = [authors.json]
            else:
                raise TypeError("authors is not of type Person")

    def menudirs(self, menudirs=None):
        """
        Set or returns default search paths for looking for menus.

        Parameter
        ----------
        menudirs: str or list of str
        When provided, set default search paths.
        """

        if menudirs is None:
            return self.config["menudirs"]

        if not isinstance(menudirs, list):
            raise TypeError("menudirs is not a list of path's")

        sysmenusdir = os.path.join(os.path.dirname(__file__),"menus")

        dirs = []

        for p in menudirs:
            dirs.append(p)

        for sub in os.listdir(sysmenusdir):
            dirs.append(os.path.join(sysmenusdir, sub))

        self.config['menudirs'] = dirs

    def save(self):
        """
        Save and apply the setup.

        The setup is save in $HOME/.config/pygebr/config.json.
        """
        with open(self.configFile,'w') as fp:
            fp.write(JSON.dumps(self.config))



#!/bin/python3

# Create menus for using with PyGêBR, that make
# common shell tools available.

from pygebr import Person, Prog, Param, Flow
import os

basepath = os.path.dirname(__file__)

# author for menus
biloti = Person(name="Ricardo Biloti",
                email="biloti@unicamp.br",
                institution="University of Campinas",
                homepage="https://www.ime.unicamp.br/~biloti")

# authors for programs
eggert = Person(name="Paul Eggert")
granlund = Person(name="Torbjorn Granlund")
haertel = Person(name="Mike Haertels")
mackenzie = Person(name="David MacKenzie")
meyering = Person(name="Jim Meyering")
rubin = Person(name="Paul Rubin")
stallman = Person(name="Richard M. Stallman")
taylor = Person(name="Ian Lance Taylor")

fsf = Person(name="Free Software Foundation, Inc.",
             homepage="https://www.fsf.org/")
#------------------------------------------------------------------------------
def commit_menu(title, description, authors, executable, tags, prog,
                fname=None):
    menu = Flow(title=title, description=description,
                authors=authors, tags=tags)

    menu.program_add(prog)

    if fname == None:
        fname = executable

    fn = os.path.join(os.path.dirname(__file__), fname + ".json")
    menu.save(fn)

#------------------------------------------------------------------------------
def cat():
    title = "Shell Cat"
    description = "Concatenate files and print on the standard output"
    executable = "cat"
    authors = biloti
    tags=['sh', 'files']

    prog = Prog(title=title, description=description,
                authors=[granlund, stallman, fsf],
                url = "https://www.gnu.org/software/coreutils",
                executable=executable,
                stdin=False, stdout=True, stderr=True)

    prog.parameter_add(Param(ptype="file", keyword=" ",
                             description="File",
                             required=True,multiple=True))

    commit_menu(title, description, authors, executable, tags, prog)

#------------------------------------------------------------------------------
def df():
    title = "Shell DF"
    description = "Report file system disk space usage"
    executable = "df"
    authors = biloti
    tags=['sh', 'file system']

    prog = Prog(title=title, description=description,
                authors=[granlund, mackenzie, eggert, fsf],
                url = "https://www.gnu.org/software/coreutils",
                executable=executable,
                stdin=False, stdout=True, stderr=True)

    prog.parameter_add(Param(ptype="flag", keyword="-a",
                             description="Include pseudo, duplicate, inaccessible file systems",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-h",
                             description="Print sizes in powers of 1024 (e.g., 1023M)",
                             default=True))

    prog.parameter_add(Param(ptype="flag", keyword="-H",
                             description="Print sizes in powers of 1000 (e.g., 1.1G)",
                             default=False))

    par = Param(ptype="enum", keyword="--block-size=",
                description="Scale factor for sizes")
    par.options([
        {"description": "KiB (1024 bytes)",   "value": "KiB"},
        {"description": "MiB (1024^2 bytes)", "value": "MiB"},
        {"description": "GiB (1024^3 bytes)", "value": "GiB"},
        {"description": "TiB (1024^4 bytes)", "value": "TiB"},
        {"description": "PiB (1024^5 bytes)", "value": "PiB"},
        {"description": "EiB (1024^6 bytes)", "value": "EiB"},
        {"description": "ZiB (1024^7 bytes)", "value": "ZiB"},
        {"description": "YiB (1024^8 bytes)", "value": "YiB"},
        {"description": "KB (1000 bytes)",   "value": "KB"},
        {"description": "MB (1000^2 bytes)", "value": "MB"},
        {"description": "GB (1000^3 bytes)", "value": "GB"},
        {"description": "TB (1000^4 bytes)", "value": "TB"},
        {"description": "PB (1000^5 bytes)", "value": "PB"},
        {"description": "EB (1000^6 bytes)", "value": "EB"},
        {"description": "ZB (1000^7 bytes)", "value": "ZB"},
        {"description": "YB (1000^8 bytes)", "value": "YB"}
    ])
    prog.parameter_add(par)

    prog.parameter_add(Param(ptype="flag", keyword="-l",
                             description="Limit listing to local file systems",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--total",
                             description="Produce a grand total",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-T",
                             description="Print type file system",
                             default=False))

    commit_menu(title, description, authors, executable, tags, prog)

#------------------------------------------------------------------------------
def du():
    title = "Shell DU"
    description = "Summarize disk usage of the set of FILEs, recursively for directories"
    executable = "du"
    authors = biloti
    tags=['sh', 'file system']

    prog = Prog(title=title, description=description,
                authors=[granlund, mackenzie, mackenzie, fsf],
                url = "https://www.gnu.org/software/coreutils",
                executable=executable,
                stdin=False, stdout=True, stderr=True)

    prog.parameter_add(Param(ptype="flag", keyword="-a",
                             description="Write counts for all files, not just directories",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--apparent-size",
                             description="Print apparent sizes, rather than disk usage",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--bytes",
                             description="Print size in bytes",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-D",
                             description="Dereference only symlinks that are listed on the command line",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-h",
                             description="Print sizes in human readable format (e.g., 1K 234M 2G)",
                             default=True))

    par = Param(ptype="enum", keyword="--block-size=",
                description="Scale factor for sizes")
    par.options([
        {"description": "KiB (1024 bytes)",   "value": "KiB"},
        {"description": "MiB (1024^2 bytes)", "value": "MiB"},
        {"description": "GiB (1024^3 bytes)", "value": "GiB"},
        {"description": "TiB (1024^4 bytes)", "value": "TiB"},
        {"description": "PiB (1024^5 bytes)", "value": "PiB"},
        {"description": "EiB (1024^6 bytes)", "value": "EiB"},
        {"description": "ZiB (1024^7 bytes)", "value": "ZiB"},
        {"description": "YiB (1024^8 bytes)", "value": "YiB"},
        {"description": "KB (1000 bytes)",   "value": "KB"},
        {"description": "MB (1000^2 bytes)", "value": "MB"},
        {"description": "GB (1000^3 bytes)", "value": "GB"},
        {"description": "TB (1000^4 bytes)", "value": "TB"},
        {"description": "PB (1000^5 bytes)", "value": "PB"},
        {"description": "EB (1000^6 bytes)", "value": "EB"},
        {"description": "ZB (1000^7 bytes)", "value": "ZB"},
        {"description": "YB (1000^8 bytes)", "value": "YB"}
    ])
    prog.parameter_add(par)

    prog.parameter_add(Param(ptype="flag", keyword="-L",
                             description="Dereference all symbolic links",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-l",
                             description="Limit listing to local file systems",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--total",
                             description="Produce a grand total",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--separate-dirs",
                             description="For directories do not include size of subdirectories",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-x",
                             description="Skip directories on different file systems",
                             default=False))

    prog.parameter_add(Param(ptype="file", keyword=" ",
                             description="Path or files to scan",
                             required=True, multiple=True))

    commit_menu(title, description, authors, executable, tags, prog)

#------------------------------------------------------------------------------
def head():
    title = "Shell Head"
    description = "Output the first part of files"
    executable="head"
    authors = biloti
    tags = ['sh', 'files']

    prog = Prog(title=title, description=description,
                authors=[mackenzie, meyering, fsf],
                url="https://www.gnu.org/software/coreutils/",
                executable=executable, stdin=True)

    prog.parameter_add(Param(ptype="integer", keyword="-n ",
              description="How many lines to output",
              default=10))

    prog.parameter_add(Param(ptype="flag", keyword="-q",
              description="Never print headers giving file names",
              default=False))

    commit_menu(title, description, authors, executable, tags, prog)

#------------------------------------------------------------------------------
def ls():
    title = "Shell LS"
    description = "List directory contents"
    executable="ls"
    authors = biloti
    tags = ['sh', 'file system']

    prog = Prog(title=title, description=description,
                authors=[stallman, mackenzie, fsf],
                url="https://www.gnu.org/software/coreutils",
                executable=executable, stdin=False)

    prog.parameter_add(Param(ptype="flag", keyword="--all",
              description="Do not ignore entries starting with .",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--author",
              description="Print the author of each file (when long format is in use)",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-d",
              description="List directories themselves, not their contents",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-F",
              description="Append indicator (one of */=>@|) to entries",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--group-directories-first",
              description="Group directories before file",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-h",
              description="Print human readable size units",
              default=True))

    prog.parameter_add(Param(ptype="flag", keyword="-l",
              description="Long list format",
              default=True))

    prog.parameter_add(Param(ptype="flag", keyword="-S",
              description="Sort list by size",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--reverse",
              description="Reverse order",
              default=False))

    prog.parameter_add(Param(ptype="path", keyword=" ",
              description="Paths to list",
              multiple=True))

    prog.parameter_add(Param(ptype="file", keyword=" ",
              description="Files to list",
              multiple=True))

    commit_menu(title, description, authors, executable, tags, prog)

#------------------------------------------------------------------------------
def sort():
    title = "Shell Sort"
    description = "Sort lines of text"
    executable="sort"
    authors = biloti
    tags = ['sh']

    prog = Prog(title=title, description=description,
                authors=[haertel, eggert, fsf],
                url="https://www.gnu.org/software/coreutils",
                executable=executable, stdin=True)

    prog.parameter_add(Param(ptype="flag", keyword="--ignore-leading-blanks",
              description="Ignore leading blanks",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--dictionary-order",
              description="Consider only blanks and alphanumeric characters",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--ignore-case",
              description="Fold lower case to upper case characters (ignore case)",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--general-numeric-sort",
              description="Compare according to general numerical value",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--ignore-nonprinting",
              description="Consider only printable characters",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--month-sort",
              description="Compare (unknown) < 'JAN' < ... < 'DEC'",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--human-numeric-sort",
              description="Compare human readable numbers (e.g., 2K 1G)",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--numeric-sort",
              description="Compare according to string numerical value",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-R",
              description="Shuffle, but group identical keys",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--reverse",
              description="Reverse the result of comparisons",
              default=False))

    prog.parameter_add(Param(ptype="string", keyword="--key=",
              description="Sort via a key (see documentation)"))

    prog.parameter_add(Param(ptype="string", keyword="--field-separator=",
              description="Field separator"))

    commit_menu(title, description, authors, executable, tags, prog)

#------------------------------------------------------------------------------
def tail():
    title = "Shell Tail"
    description = "Output the last part of files"
    executable="tail"
    authors = biloti
    tags = ['sh','files']

    prog = Prog(title=title, description=description,
                authors=[rubin, mackenzie, taylor, meyering, fsf],
                url="https://www.gnu.org/software/coreutils/",
                executable=executable, stdin=True)

    prog.parameter_add(Param(ptype="flag", keyword="-f",
              description="Output appended data as the file grows",
              default=False))

    prog.parameter_add(Param(ptype="integer", keyword="-n ",
              description="How many lines to output",
              default=10))

    commit_menu(title, description, authors, executable, tags, prog)

#------------------------------------------------------------------------------
def uniq():
    title = "Shell Uniq"
    description = "Report or omit repeated lines"
    executable="uniq"
    authors = biloti
    tags = ['sh','files']

    prog = Prog(title=title, description=description,
                authors=[stallman, mackenzie, fsf],
                url="https://www.gnu.org/software/coreutils/",
                executable=executable, stdin=True)

    prog.parameter_add(Param(ptype="flag", keyword="--count",
              description="Prefix lines by the number of occurrences",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--repeated",
              description="Only print duplicate lines, one for each group",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-D",
              description="Print all duplicate lines",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--ignore-case",
              description="Ignore differences in case when comparing",
              default=False))

    prog.parameter_add(Param(ptype="flag", keyword="--uniq",
              description="Only print unique lines",
              default=False))

    prog.parameter_add(Param(ptype="integer", keyword="--skip-fields=",
              description="Number of first fields to skip when comparing"))

    prog.parameter_add(Param(ptype="integer", keyword="--skip-chars=",
              description="Number of characters to skip when comparing"))

    prog.parameter_add(Param(ptype="integer", keyword="--check-chars=",
              description="Maximum number of characters in lines to compare"))


    commit_menu(title, description, authors, executable, tags, prog)

#------------------------------------------------------------------------------
def wc():
    title = "Shell WC"
    description = "Print newline, word, and byte counts for a file"
    executable="wc"
    authors = biloti
    tags = ['sh','files']

    prog = Prog(title=title, description=description,
                authors=[rubin, mackenzie, fsf],
                url="https://www.gnu.org/software/coreutils/",
                executable=executable, stdin=True)

    prog.parameter_add(Param(ptype="flag", keyword="-c",
                             description="Count bytes",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-m",
                             description="Count characters",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-l",
                             description="Count newlines",
                             default=True))

    prog.parameter_add(Param(ptype="flag", keyword="-w",
                             description="Count words",
                             default=False))

    prog.parameter_add(Param(ptype="flag", keyword="-L",
                             description="Print the maximum display width",
                             default=False))

    commit_menu(title, description, authors, executable, tags, prog)

#------------------------------------------------------------------------------
cat()
df()
du()
head()
ls()
sort()
tail()
uniq()
wc()



# This file is part of PyGeBR library.
#
# PyGeBR is a library to create, merge, edit, inspect and run
# processing flows. It was originally designed for seismic
# data processing, but it extends to other data as well.
#
# PyGeBR is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyGeBR is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyGeBR.  If not, see <https://www.gnu.org/licenses/>
#
# Copyright 2021-2022 Ricardo Biloti <biloti@unicamp.br>
#                     Eduardo Filpo <efilpo@gmail.com>
#

import functools
import ipywidgets as W
import re

#-------------------------------------------------------------------------------
class Param:
    """Class used to represent a program parameter.

A program parameter can be of one of these types:
flag:
    a parameter that may or may not be present in command line

integer:
    a parameter that expects an integer value

float:
    a parameter that expects an float value

range:
    a parameter that expects value within an acceptable range

string:
    a parameter that expects an string

integers:
    a parameter that expects an array of integer values

floats:
    a parameter that expects an array of float values

strings:
    a parameter that expects an array of string values

file:
    a parameter that expects an file name

path:
    a parameter that expects a path

enum:
    a parameter that expects a value from an list of acceptable options

section:
    a metaparameter used only to group subsequent parameters inside section widget

comment:
    a metaparameter used only to add a line of comment in
    parameters list widget


Literal or expression
---------------------

A parameter's value can be represented by a literal value or by an
expression that has to be evaluated. An expression can employ variables
defined in global and local scopes.

Attributes
----------

json: dict
    Dictionary representing a parameter

widget: dict
    Dictionary holding an widget to edit parameter's attributes

type: str
    Parameter's type

keyword: str
    Keyword in command line used to set the parameter

description: str
    One-line description for the parameter purpose

value: (depending on parameter's type)
    Parameter's value

default: (depending on parameter's type)
    Paramter's default value used to fillin widget

required: boolean
    Flag to indicate whether the parameter is required by the program

multiple: boolean
    Flat to indicate whether the parameter can be provided multiple
    times in command line.

Below, in setter methods, providing "" as value has the effect to delete
the attribute. Note that a parameter has to have its type defined.
"""

    _boolean_types = ['flag']
    _scalar_types = ['enum','file','float','integer','path','range','string']
    _array_types = ['floats','integers','strings']
    _meta_types = ['section','comment']

    _supported_types = _boolean_types + _scalar_types + _array_types + _meta_types

    def __init__(self, json=None, ptype=None, keyword=None, description=None,
                 required=False, multiple=False, separator=",", title=None,
                 placeholder=None, default=None, value=None):
        """
        Parameters
        ----------

        json: dictionary (optional)
            A dictionary to represent a person.

        ptype: str (optional)
            Parameter type.
            Acceptable values are

            'comment', 'enum', 'file', 'flag', 'float', 'floats', 'integer',
            'integers', 'path', 'range', 'section', 'string', or 'strings'.

        keyword: str (optional)
            Keyword to provide set the parameter in command line

        description: str (optional)
            One-line sentence describing the parameter's purpose

        required: boolean
            Flag indicating whether the parameter is required by the program

        multiple: boolean
            Flat to indicate whether the parameter can be provided multiple
            times in command line

        separator: str (optional)
            String used as separator between elements of arrays

        title: str (optional)
            Title for section-type parameter in widget

        placeholder: str (optional)
            Placeholder for parameter's entry in widget

        default: (optinal)
            Default value to fill in UI
        """

        if json is None:
            self.json = {}
        else:
            if not isinstance(json, dict):
                raise TypeError("Argument should be a dictionary")

            self.json = json
            required = None
            multiple = None

        # Widget
        self.widget = {}

        if ptype is not None:
            self.type(ptype)

        if 'type' not in self.json:
            raise Exception("Missing parameter type")

        if ptype == 'section':
            required = None
            multiple = None
            if title is not None:
                self.title(title)
            return

        if ptype == 'comment':
            required = None
            multiple = None
            if description:
                self.description(description)
            return

        if title is not None:
            raise Exception("Only section-type parameters admit title attribute")

        # Flag parameters can be provided only once
        if ptype == "flag":
            multiple = False
            required = None

        if keyword is not None:
            self.keyword(keyword)

        if description is not None:
            self.description(description)

        if default is not None:
            self.default(default)

        if value is not None:
            self.values(value)

        if required is not None:
            self.required(required)

        if multiple is not None:
            self.multiple(multiple)

        if (ptype in ['string','file','path'] + self._array_types):
            if placeholder is not None:
                self.placeholder(placeholder)
            elif ptype in self._array_types:
                self.placeholder("List of %s, separated by \"%s\""%(ptype, separator))

        if ptype in self._array_types:
            self.separator(separator)

    def __str__(self):
        """
        Return a string with the parameter as it would appear in
        command line, without resolving expressions though. If the
        parameter hasn't value set, nothing is returned.

        To resolve expressions use self.eval().
        """

        text = ""
        # If parameter has no value set returns nothing
        if not self.has_non_empty_value():
            return ""

        if ((len(self.values()) >1) and not self.multiple()):
            raise Exception("Parameter %s defined more than once"%self.keyword())

        for value in self.values():
            if self.type()  == "flag":
                if 'expression' in value:
                    raise Exception("Expression for flags not supported")

                if value['literal']:
                    text = " ".join([text,self.keyword()]).strip()
            else:
                if 'literal' in value:
                    valuestr = self._eval_literal(value['literal'])
                    text = " ".join([text,self.keyword() + "\"" + valuestr + "\""]).strip()
                else:
                    if value['expression']:
                        valuestr = str(value['expression'])
                        text = " ".join([text,self.keyword() + "\"" + valuestr + "\""]).strip()

        return text.strip()

    def eval(self, params=None):
        """
        Return a string with the parameter as it would appear in
        command line, resolving expressions. If the
        parameter hasn't value set, nothing is returned.

        Parameter
        ---------

        params: dict (optional)
           Dictionary of variables to define local scope.
        """

        text = ""
        # If parameter has no value set returns nothing
        if not self.has_non_empty_value():
            return ""

        if ((len(self.values()) >1) and not self.multiple()):
            raise Exception("Parameter '%s' defined more than once"%self.description())

        if self.type() == "string":
            if not self.match_regexp():
                raise ValueError("At least on literal value for parameter '%s' isn't valid against acceptable regexp '%s'"%(self.description(),self.regexp()))

        for value in self.values():

            # Treatment for flags
            if self.type()  == "flag":
                if 'expression' in value:
                    raise Exception("Expression for flags isn't supported")

                if value['literal']:
                    text = " ".join([text,self.keyword()]).strip()
                continue

            # Treatment for literal values
            if 'literal' in value:
                if self.type() == "integer" or (self.type() == 'range' and self.json['digits'] == 0):
                    valuestr = str("%i"%value['literal'])
                else:
                    valuestr = str(value['literal'])

                text = " ".join([text,self.keyword() + "\"" + valuestr + "\""]).strip()
                continue

            # Treatment for non-arrays expressions
            if self.type() in self._scalar_types:

                try:
                    valuestr = self._eval_expression(value['expression'], params)
                except:
                    raise

                text = " ".join([text,self.keyword() + "\"" + valuestr + "\""]).strip()
                continue

            # Treatment for expressions in arrays
            valueliststr = self._eval_array(value['expression'], params)
            text = " ".join([text,self.keyword() + "\"" + valueliststr + "\""]).strip()
            continue

        return text.strip()

    def type(self, value=None):
        """
        Query or set parameter's type

        Parameter
        ---------

        value: str (optinal)
            value to set parameter's value
        """
        if value is None:
            return self.json['type']

        if value not in self._supported_types:
            raise Exception("Unknown parameter type")

        self.json['type'] = value

    def has_keyword(self):
        """
        Return whether a parameter has keyword set
        """
        return 'keyword' in self.json

    def keyword(self, value = None):
        """
        Query or set parameter's keyword

        Parameter
        ---------

        value: str (optinal)
            value to set parameter's keyword
        """
        if value is None:
            return self.json['keyword'] if self.has_keyword() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if not value:
            if self.has_keyword():
                del self.json['keyword']
        else:
            self.json['keyword'] = value

    def has_description(self):
        """
        Return whether a parameter has description set
        """
        return 'description' in self.json

    def description(self, value = None):
        """
        Query or set parameter's description

        Parameter
        ---------

        value: str (optinal)
            value to set parameter's description
        """
        if value is None:
            return self.json['description'] if self.has_description() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if not value:
            if self.has_description():
                del self.json['description']
        else:
            self.json['description'] = value

    def has_required(self):
        """
        Return whether a parameter has required attribute set
        """
        return 'required' in self.json

    def required(self, value = None):
        """
        Query or set parameter's required attribute

        Parameter
        ---------

        value: str or boolean (optional)
            value to set parameter's required flag

        """
        if value is None:
            return self.json['required'] if self.has_required() else False

        if isinstance(value, str) and (not value) and self.has_required():
            del self.json['required']
            return

        if not isinstance(value, bool):
            raise TypeError("Argument should be a boolean")

        self.json['required'] = value

    def has_multiple(self):
        """
        Return whether a parameter has multiple attribute set
        """
        return 'multiple' in self.json

    def multiple(self, value = None):
        """
        Query or set parameter's multiple attribute

        Parameter
        ---------

        value: str or boolean (optional)
            value to set parameter's multiple flag

        """
        if value is None:
            return self.json['multiple'] if self.has_multiple() else False

        if isinstance(value, str) and (not value) and self.has_multiple():
            del self.json['multiple']
            return

        if not isinstance(value, bool):
            raise TypeError("Argument should be a boolean")

        self.json['multiple'] = value

    def has_separator(self):
        """
        Return whether a parameter has separator set
        """
        return 'separator' in self.json

    def separator(self, value=None):
        """
        Query or set parameter's separator value

        Parameter
        ---------

        value: str (optinal)
            separator that goes between array elements
        """
        if not self.type() in ["integers", "floats", "strings"]:
            raise Exception("Parameter type doesn't support separator")

        if value is None:
            return self.json['separator'] if self.has_separator() else None

        if isinstance(value, str) and (not value):
            if self.has_separator():
                del self.json['separator']
        else:
            self.json['separator'] = value

    def has_default(self):
        """
        Return whether a parameter has default set
        """
        return 'default' in self.json

    def default(self, value=None):
        """
        Query or set parameter's default value

        Parameter
        ---------

        value: str (optinal)
            Value to set parameter's widget entry default.
            Provide an empty string to delete the default.
        """
        if value is None:
            return self.json['default'] if self.has_default() else None

        # Empty string, deletes the default
        if isinstance(value, str) and (not value):
            if self.has_default():
                del self.json['default']
            return

        if self.type() == "integer":
            if not isinstance(value, int):
                raise TypeError("Default value isn't an integer")
            else:
                self.json['default'] = value
                return

        if self.type() in ["range", "float"]:
            if not (isinstance(value, float) or isinstance(value,int)):
                raise TypeError("Default value isn't a number")
            else:
                self.json['default'] = value
                return

        if self.type() in ["string", "enum", "file", "path"]+self._array_types:
            if not isinstance(value, str):
                raise TypeError("Default value isn't a string")
            else:
                self.json['default'] = value
                return

        if self.type() == "flag":
            if not isinstance(value, bool):
                raise TypeError("Default value isn't True or False")
            else:
                self.json['default'] = value
                return

        if self.type() in self._meta_types:
            raise Exception("Meta parameter hasn't default attribute")

        self.json['default'] = value

    def has_value(self):
        """
        Return whether a parameter has value set
        """
        return 'value' in self.json

    def has_non_empty_value(self):
        """
        Return whether a parameter has value set
        and value isn't empy
        """
        if not self.has_value():
            return False

        for c in self.value_content():
            if len(str(c)) > 0:
                return True

        return False

    def values(self, value=None):
        """
        Query or set parameter's values

        Parameter
        ---------

        value: (optinal)
            value to set parameter's value.

            value can be:
            - A single literal value, or
            - A list of literal/expression values

            See these examples

            # Set an integer parameter value to 12
            intpar.values(12)

            # Also set an integer parameter value to 12
            intpar.values([{'literal': 12}])

            # Set an integer parameter value to 2 * N,
            # where N is a variable defined in the global or local scopes.
            intpar.values([{'expression': '2 * N'}])

            # Set an integer parameter value for a parameter
            # that accepts a list of values (multiple = True)
            intpar.values([{'expression': '2 * N'}, {'literal': 12}])

            # Set a string parameter
            strpar.values([ {'literal': 'some text'}, {'expression': '\'My name is\' + name'}])

            # All array-type parameters has to be set as strings
            intspar.values("2;4;5")
            intspar.values([ {'expression': '2;4;5+7'} ])
        """
        if value is None:
            return self.json['value'] if self.has_value() else []

        if isinstance(value, str) and (not value):
            if self.has_value():
                del self.json['value']
        elif isinstance(value, dict):
            self.json['value'] = [value]
        elif isinstance(value, list):
            self.json['value'] = value
        else:
            self.json['value'] = [{"literal": value}]

    def value_content(self):
        """
        Query parameter's value content
        """
        content = []
        for value in self.values():
            if 'literal' in value:
                content.append(value['literal'])
            else:
                content.append(value['expression'])

        return content

    def has_range(self):
        """
        Return whether a range parameter has range set
        """
        if self.type() != "range":
            raise Exception("Parameter type isn't range")

        return 'min' in self.json and 'max' in self.json

    def range(self, limits=None, vinc=None, vdigits=None):
        """
        Query or set parameter's range

        Parameters
        ----------

        limits: list (optional)
            [lower limit value, upper limit value]

        vinc: float (optional)
            value increment in range entry widget

        vdigits: integer (optional)
            amount of digits to display in range entry widget

        """
        if limits is None:
            return [self.json['min'],self.json['max']] if self.has_range() else None

        if not isinstance(limits, list):
            raise TypeError("Argument should be a list of two integers")

        self.json['min'] = limits[0]
        self.json['max'] = limits[1]
        if vinc is not None:
            self.json['inc'] = vinc
        if vdigits is not None:
            self.json['digits'] = vdigits

    def has_regexp(self):
        """
        Return whether a string parameter has regexp set
        """

        return 'regexp' in self.json

    def regexp(self, value=None):
        """
        Query or set string parameter's regexp

        Parameter
        ---------

        value: str (optinal)
            regular expression a string parameter has to be validated against
        """
        if self.type() != "string":
            raise Exception("Parameter's type isn't string")

        if value is None:
            return self.json['regexp'] if self.has_regexp() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if (not value):
            if self.has_regexp():
                del self.json['regexp']
        else:
            self.json['regexp'] = value

    def match_regexp(self):
        """
        Check if a literal value matches with the
        regexp, for string-type parameter.

        Parameter
        ---------

        value: str
           String to match with regular expression

        Return
        ------
           True if value matches with regular expression
           or if there is no regular expression.
        """

        if self.type() != "string":
            return False

        if not self.has_regexp():
            return True

        for value in self.values():
            if 'literal' in value:
                if re.match(self.regexp(),value['literal']) == None:
                    return False

        return True

    def has_fileType(self):
        """
        Return whether a file parameter has type set
        """

        return 'fileType' in self.json

    def fileType(self, value=None):
        """
        Query or set parameter's file type

        Parameter
        ---------

        value: (optinal)
            value to set file type in entry widget for file-type parameters
        """
        if self.type() != "file":
            raise Exception("Parameter's type isn't file")

        if value is None:
            return self.json['fileType'] if self.has_fileType () else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if (not value):
            if self.has_fileType():
                del self.json['fileType']
        else:
            self.json['fileType'] = value

    def has_filePattern(self):
        """
        Return whether a file parameter has pattern set
        """

        return 'filePattern' in self.json

    def filePattern(self, value=None):
        """
        Query or set parameter's file pattern

        Parameter
        ---------

        value: (optinal)
            value to set file pattern in entry widget for file-type parameters
        """
        if self.type() != "file":
            raise Exception("Parameter's type isn't file")

        if value is None:
            return self.json['filePattern'] if self.has_filePattern () else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if not value:
            if self.has_filePattern():
                del self.json['filePattern']
        else:
            self.json['filePattern'] = value

    def has_title(self):
        """
        Return whether a section parameter has title set
        """
        return 'title' in self.json

    def title(self, value=None):
        """
        Query or set section parameter's title

        Parameter
        ---------

        value: (optinal)
            value to set section's title
        """
        if self.type() != "section":
            raise Exception("Parameter's type isn't section")

        if value is None:
            return self.json['title'] if self.has_title () else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if not value:
            if self.has_title():
                del self.json['title']
        else:
            self.json['title'] = value

    def has_placeholder(self):
        """
        Return whether a string parameter has placeholder set
        """

        return 'placeholder' in self.json

    def placeholder(self, value=None):
        """
        Query or set string parameter's placeholder

        Parameter
        ---------

        value: (optinal)
            value to set placeholder string in widget entry
        """
        if not self.type() in ["string", "file", "path"] + self._array_types:
            raise Exception("Parameter's type isn't string")

        if value is None:
            return self.json['placeholder'] if self.has_placeholder () else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if not value:
            if self.has_placeholder():
                del self.json['placeholder']
        else:
            self.json['placeholder'] = value

    def has_options(self):
        """
        Return whether a enum parameter has options set
        """
        if self.type() != "enum":
            raise Exception("Parameter's type isn't enum")

        return 'options' in self.json

    def options(self, value=None):
        """
        Query or set enum parameter's options

        Parameter
        ---------

        value: list of dict (optional)
            List of acceptable options for enum-type parameter

            This list has the format:
            [
              {"description": "Sunny season", "value": "summer"},
              {"description": "Colder season", "value": "winter"},
              {"description": "Flower's season", "value": "spring"},
              {"description": "Sadiest season", "value": "autumn"}
            ]
        """
        if self.type() != "enum":
            raise Exception("Parameter's type isn't enum")

        if value is None:
            return self.json['options'] if self.has_options () else []

        if not isinstance(value, list):
            raise TypeError("Argument should be a list")

        if not value:
            if self.has_options():
                del self.json['options']
        else:
            self.json['options'] = value

    def formated(self, isfull, resolve, pardict):
        """
        Return a list of strings describing the parameter

        Parameter
        ---------

        isfull: boolean
            True for a more verbose output

        resolve: boolean
            True to evaluate expressions in parameter's value

        pardict: dictionary
            Dictionary to define local scope to evaluate expressions
        """

        if self.type() in self._scalar_types:
            return self._dump_par(isfull, resolve, pardict)

        if self.type() in self._boolean_types:
            return self._dump_flag_par(isfull, resolve, pardict)

        if self.type() in self._array_types:
            return self._dump_array_par(isfull, resolve, pardict)

        if self.type() in self._meta_types:
            if self.type() == 'section':
                return self.title()

            if self.type() == 'comment':
                return self.description()

        return ["Unknown parameter type"]

    def w_line(self):
        """
        Return a widget to set/acquire the parameter

        The widget is meant to be included into a VBox.
        """
        if self.widget == {}:
            self.__widget_construct()

        return self.widget['line']

    def w_set(self):
        """
        Set widget value from parameter's value

        Note: Doesn't support flags, ranges and enums with expressions.
        """

        if self.type() in ["flag", "range", "enum"]:
            if self.has_value():
                self.widget['entry'][0].value = self.values()[0]['literal']
            elif self.has_default():
                self.widget['entry'][0].value = self.default()
            return

        if self.type() in self._meta_types:
            return

        # Set for parameters that can hold expressions
        if self.has_value():

            values = self.values()

            for k in range(len(values)):
                value = values[k]
                if k == len(self.widget['entry']):
                    on_add_clicked(button = None,par=self)

                if 'expression' in value:
                    self.widget['exp'][k].value = True
                    self.widget['expression'][k].value = value['expression']
                else:
                    self.widget['exp'][k].value = False
                    self.widget['literal'][k].value = value['literal']

            while len(values) < len(self.widget['entry']):
                on_del_clicked(button = None,par=self)

        elif self.has_default():
            for k in range(len(self.widget['entry'])):
                self.widget['exp'][k].value = False
                self.widget['literal'][k].value = self.default()
        else:
            for k in range(len(self.widget['entry'])):
                if self.type() in ['integer','float']:
                    self.widget['exp'][k].value = True
                    self.widget['expression'][k].value = ""
                else:
                    self.widget['exp'][k].value = False
                    self.widget['literal'][k].value = ""

        ninstances = len(self.widget['entry'])
        for k in range(ninstances):
            if self.widget['exp'][k].value:
                self.widget['entry'][k] = self.widget['expression'][k]
                self.widget['expression'][k].layout.display = 'flex'
                self.widget['literal'][k].layout.display = 'none'
            else:
                self.widget['entry'][k] = self.widget['literal'][k]
                self.widget['expression'][k].layout.display = 'none'
                self.widget['literal'][k].layout.display = 'flex'


    def w_get(self):
        """
        Get parameter's value from widget value
        """
        if self.type() in self._meta_types:
            return

        nentries = len(self.widget['entry'])

        k = 0
        values = []
        empty = True
        for k in range(nentries):
            value = self.widget['entry'][k].value
            if 'exp' in self.widget and self.widget['exp'][k].value:
                empty = False
                values.append({"expression": value})
            else:
                if self.type() == 'integer':
                    values.append({"literal": int(value)})
                    empty = False
                elif self.type() in ['float', 'range']:
                    values.append({"literal": float(value)})
                    empty = False
                else:
                    if isinstance(value, bool) or (isinstance(value, str) and len(value) > 0):
                        empty = False
                    values.append({"literal": value})

        if empty:
            self.values("")
        else:
           self.values(values)


    def _eval_literal(self, lit):
        """
        Eval a literal (scalar or array) to assure it is correctly defined

        Parameter
        ---------

        lit:
           Literal to evaluate

        Return
        ------
           An string representing the literal value
        """

        if self.type() == "integer" or (self.type() == 'range' and self.json['digits'] == 0):
            try:
                valuestr = "%i"%int(lit)
            except TypeError:
                raise TypeError(str(lit) + " doesn't cast to integer, in parameter '" + self.description() + "'")
            except ValueError:
                raise ValueError(str(lit) + " isn't an integer, in parameter '" + self.description() + "'")
            except Exception as err:
                raise Exception(str(err))

            return valuestr

        if self.type() == "integers":
            values = lit.split(self.separator())
            valuestrlist = ""
            for value in values:
                try:
                    valuestr = "%i"%int(value)
                except TypeError:
                    raise TypeError(value + " doesn't cast to integer, in parameter '" + self.description() + "'")
                except ValueError:
                    raise ValueError(value + " isn't an integer, in parameter '" + self.description() + "'")
                except Exception as err:
                    raise Exception(str(err))

                valuestrlist = self.separator().join([valuestrlist,valuestr]).strip(self.separator())

            return valuestrlist

        return str(lit)


    def _eval_expression(self, exp, params):
        """
        Eval an expression

        Parameters
        ----------

        exp: str
           String with expressions separeted by self.separator()

        params: dict
           Dictionary of variables to define local scope.
        """

        try:
            valuestr = str(eval(exp,globals(),params))

        except NameError as err:
            errormsg = str(err) + " in expression '" + exp + \
                "', for parameter '" + self.description() + "'"
            if params == None:
                errormsg = errormsg + "\n (indeed, a dictionary of parameters wasn't even provided)"
            raise NameError(errormsg)

        except SyntaxError as err:
            raise SyntaxError("syntax error in expression '" + exp +
                              "', for parameter '" + self.description() + "'" )

        except ZeroDivisionError:
            raise ZeroDivisionError("division by zero while evaluating expression '" +
                                    exp + "', for parameter '" +
                                    self.description() + "'" )
        except:
            raise Exception("unable to evaluate '" + exp + "' in parameter '" + self.description() + "'")

        if self.type() == 'integer':
            try:
                vint = int(float(valuestr))
            except ValueError as err:
                raise ValueError("expression '" + exp + "' doesn't cast to integer, for parameter '" +
                                 self.description()+"'")
            except:
                raise("unable to eval expression'" + exp + "' doesn't cast to integer, for parameter '" +
                                 self.description()+"'")

            valuestr = f"{vint}"

        return valuestr

    def _eval_array(self, exps, params):
        """
        Eval an array of expressions

        Parameters
        ----------

        exps: str
           String with expressions separeted by self.separator()

        params: dict
           Dictionary of variables to define local scope.
        """

        valuelist = exps.split(self.separator())
        valueliststr = ""

        for vexp in valuelist:

            valuestr = self._eval_expression(vexp, params)
            valueliststr = self.separator().join([valueliststr,valuestr]).strip(self.separator())

        return valueliststr

    #----------------------------------------------------------------------------
    def _dump_comment(self):
        """
        Private function to generate the comment concerning
        a parameter (its type, default values and attributes)
        """

        output = "[" + self.type()
        if self.type() in ['integers', 'floats', 'strings']:
            output = output + "(" + self.separator() + ")"

        if self.type() == 'range':
            template = "%." + str(self.json['digits']) + "f"
            output = output + ", " + template%self.json['min']+":" + \
                     template%self.json['inc']+":" + template%self.json['max']

        if self.has_default():
            if self.type() == 'flag':
                if self.default():
                    output = output + ", on]"
                else:
                    output = output + ", off]"
            else:
                output = output + ", " + str(self.default()) + "]"
        else:
            output = output + "]"

        if self.required():
            output = output + " R"

        if self.multiple():
            output = output + " M"

        return output

    #--------------------------------------------------------------------------------
    # Private methods for dump parameters

    def _dump_par(self, isfull, resolve, pardict):
        """
        Private function to dump a number or string parameter
        """

        comment = self._dump_comment()

        output=[]
        keyvalue = self.keyword()
        if self.has_non_empty_value():
            values = self.values()
            first = True
            for value in values:
                valuestr = ""
                if "literal" in value:
                    valuestr = self._eval_literal(value['literal'])
                else:
                    if resolve:
                        valuestr = self._eval_expression(value['expression'], pardict)
                    else:
                        valuestr = value['expression']

                line = keyvalue+"\""+valuestr+"\""
                if first:
                    output.append("%-40s"%line + " %s"%comment)
                    first = False
                else:
                    output.append("%-40s"%line)
        else:
            output.append("%-40s"%keyvalue + " %s"%comment)

        if isfull:
            output.append("%-70s"%self.description())
            if self.has_regexp():
                output.append("%-70s"%"Regular expression to match literal values:")
                output.append("%-70s"%self.regexp())
            if self.type() == 'enum':
                output.append("Valid options are:")
                options = self.json['options']
                for opt in options:
                    output.append("%15s: %s"%(opt['value'], opt['description']))

        return output

    #--------------------------------------------------------------------------------
    def _dump_array_par(self, isfull, resolve, pardict):
        """
        Private function to dump an array parameter
        """

        comment = self._dump_comment()

        output=[]
        keyvalue = self.keyword()
        if self.has_non_empty_value():
            values = self.values()
            first = True
            for value in values:
                valuestr = ""
                if "literal" in value:
                    valuestr = self._eval_literal(value['literal'])
                else:
                    if resolve:
                        valuestr = self._eval_array(value['expression'], pardict)
                    else:
                        valuestr = value['expression']

                line = keyvalue+"\""+valuestr+"\""
                if first:
                    output.append("%-40s"%line + " %s"%comment)
                    first = False
                else:
                    output.append("%-40s"%line)
        else:
            output.append("%-40s"%keyvalue + " %s"%comment)

        if isfull:
            output.append("%-70s"%self.description())
            if self.type() == 'enum':
                output.append("Valid options are:")
                options = self.json['options']
                for opt in options:
                    output.append("%15s: %s"%(opt['value'], opt['description']))

        return output

    #--------------------------------------------------------------------------------
    def _dump_flag_par(self, isfull, resolve, pardict):
        """
        Private function to dump a flag parameter
        """

        comment = self._dump_comment()

        output = []
        if self.has_value():
            value = self.values()[0]
            if "literal" in value:
                if value['literal']:
                    valuestr = " (on)"
                else:
                    valuestr = " (off)"
            else:
                if resolve:
                    try:
                        valuestr = str(eval(value['expression'],globals(),pardict))
                    except:
                        raise Exception("Unable to evaluate parameter")
                else:
                    valuestr = value['expression']

        else:
            valuestr = " (off)"

        line = self.keyword() + valuestr
        output.append("%-40s"%line + " %s"%comment)

        if isfull:
            output.append("%-70s"%self.description())

        return output


    #--------------------------------------------------------------------------------
    # Private methods for widget construction

    def __widget_construct(self):
        """
        Construct the widget depending of the parameter's type
        """

        if not self.multiple():
            if self.type() == 'integer':
                self.__construct_w_meta_param(W.IntText)
            elif self.type() == 'float':
                self.__construct_w_meta_param(W.FloatText)
            elif self.type() in ['string','file','path', 'strings', 'integers','floats']:
                self.__construct_w_meta_param(W.Text)
            elif self.type() == 'flag':
                self.__construct_w_flag()
            elif self.type() == 'range':
                self.__construct_w_range()
            elif self.type() == 'enum':
                self.__construct_w_enum()
            elif self.type() == 'comment':
                self.__construct_w_comment()
            else:
                self.__construct_w_section()
        else:
            if self.type() in ["string","file","path"]:
                self.__construct_w_meta_param_multiple(W.Text)
            elif self.type() == "integer":
                self.__construct_w_meta_param_multiple(W.IntText)
            elif self.type() == "float":
                self.__construct_w_meta_param_multiple(W.FloatText)
            else:
                raise Exception("No support for multiple instances of %s parameter"%self.type())


    def __construct_w_meta_param(self,Entry_widget):
        """
        Widget for (integer,float,string,file,path)-type parameter
        which can't be instanciate.
        """
        label = W.Label(value = self.description(), layout = W.Layout(width="60%"))
        label_set_class(self.required(), label, "required")

        self.widget['literal'] = [ Entry_widget(layout = W.Layout(width="35%")) ]
        self.widget['exp'] = []
        self.widget['entry'] = []
        self.widget['expression'] = []

        self.__init_widget(0)

        self.widget['line'] = W.HBox([label,
                                      self.widget['literal'][0],
                                      self.widget['expression'][0],
                                      self.widget['exp'][0]])


    def __construct_w_meta_param_multiple(self,Entry_widget):
        """
        Widget for (integer,float,string,file,path)-type parameter
        that can be provided multiple times
        """
        label = W.Label(value = self.description())
        label_set_class(self.required(), label, "required")

        bt_layout = W.Layout(indent=False,width="30px")
        bt_add = W.Button(icon="plus", tooltip="Add an instance", layout = bt_layout)
        bt_del = W.Button(icon="minus", tooltip="Remove last instance", layout = bt_layout)
        bt_add.style.button_color = 'white'
        bt_del.style.button_color = 'white'

        bt_add.on_click(functools.partial(on_add_clicked, par=self))
        bt_del.on_click(functools.partial(on_del_clicked, par=self))
        label_and_buttons = W.HBox([label, bt_add, bt_del], layout = W.Layout(width="60%"))

        self.widget['entry_widget'] = Entry_widget

        values = self.values()
        self.widget['ninstances'] = max(1,len(values))

        self.widget['literal'] = []
        self.widget['exp'] = []
        self.widget['entry'] = []
        self.widget['expression'] = []

        children = ()
        for k in range(self.widget['ninstances']):

            self.widget['literal'].append(Entry_widget())

            self.__init_widget(k)
            self.widget['expression'][k].layout.width = "100%"

            hbox = W.HBox([self.widget['literal'][k],
                           self.widget['expression'][k],
                           self.widget['exp'][k]], layout=W.Layout(width="100%"))
            children = children + (hbox,)

        self.widget['vbox'] = W.VBox(children, layout=W.Layout(width="39%",height="100%"))
        self.widget['line'] = W.HBox([label_and_buttons, self.widget['vbox']])


    def __construct_w_range(self):
        """
        Widget for range-type parameter
        """
        label = W.Label(value = self.description(), layout = W.Layout(width="60%"))
        label_set_class(self.required(), label, "required")


        if 'digits' in self.json and self.json['digits'] > 1:
            self.widget['literal'] = [W.BoundedFloatText(layout = W.Layout(width="35%"),
                                min = self.json['min'],
                                max = self.json['max'],
                                step = self.json['inc'])]
        else:
            self.widget['literal'] = [W.BoundedIntText(layout = W.Layout(width="35%"),
                                min = self.json['min'],
                                max = self.json['max'],
                                step = self.json['inc'])]

        self.widget['exp'] = []
        self.widget['entry'] = []
        self.widget['expression'] = []

        self.__init_widget(0)

        self.widget['line'] = W.HBox([label,
                                      self.widget['literal'][0],
                                      self.widget['expression'][0],
                                      self.widget['exp'][0]])

    def __construct_w_enum(self):
        """
        Widget for enum-type parameter
        """
        label = W.Label(value = self.description(), layout = W.Layout(width="60%"))
        label_set_class(self.required(), label, "required")

        self.widget['entry'] = [W.Dropdown(layout = W.Layout(width="40%"))]

        options = []
        if self.has_options():
            if not self.required():
                options.append( ('-- unset --', '') )
            for opt in self.options():
                options.append( (opt['description'],opt['value']) )
        self.widget['entry'][0].options = options

        if self.has_value():
            self.widget['entry'][0].value = self.value_content()[0]
        elif self.has_default():
            self.widget['entry'][0].value = self.default()

        self.widget['line'] = W.HBox([label, self.widget['entry'][0]])

    def __construct_w_flag(self):
        """
        Widget for flag-type parameter
        """

        self.widget['entry'] = [ W.Checkbox(description=self.description(),
                                            indent=False,
                                            layout = W.Layout(width="100%")) ]

        if self.has_value():
            self.widget['entry'][0].value = self.value_content()[0]
        elif self.has_default():
            self.widget['entry'][0].value = self.default()
        self.widget['line'] = W.HBox([self.widget['entry'][0]])

    def __construct_w_section(self):
        """ Widget for section-type parameter
        """
        vbox = W.VBox()
        self.widget['line'] = W.Accordion(selected_index=None,
                                          children=(vbox,))
        self.widget['line'].set_title(0,self.title())


    def __construct_w_comment(self):
        """ Widget for comment-type parameter
        """
        self.widget['line'] = W.HTML(value='<div class="commentpar">'+self.description()+'</div>',
                                     layout=W.Layout(width="98%"))

    def __init_widget(self, index):
        """
        Initialize a parameter's widget that has support
        to expressions
        """

        if self.has_placeholder():
            self.widget['literal'][index].placeholder = self.placeholder()

        self.widget['exp'].append (W.Checkbox(indent=False, layout=W.Layout(width="25px")))
        self.widget['expression'].append (W.Text(layout = W.Layout(width="35%"),
                                                 placeholder="expression to evaluate"))

        if self.has_value() and index < len(self.values()):
            value = self.values()[index]
            if 'literal' in value:
                self.widget['exp'][index].value = False
                self.widget['literal'][index].value = value['literal']
            else:
                self.widget['exp'][index].value = True
                self.widget['expression'][index].value = value['expression']
        elif self.has_default():
            self.widget['exp'][index].value = False
            self.widget['literal'][index].value = self.default()
        else:
            # No value nor default? Set it as an empty expression
            self.widget['exp'][index].value = True

        # Which widget to show?
        if self.widget['exp'][index].value:
            self.widget['entry'].append(self.widget['expression'][index])
            self.widget['literal'][index].layout.display = 'none'
            self.widget['expression'][index].layout.display = 'flex'
        else:
            self.widget['entry'].append(self.widget['literal'][index])
            self.widget['literal'][index].layout.display = 'flex'
            self.widget['expression'][index].layout.display = 'none'

        # From now on, start observe changes in self.widget['exp'][0].value
        self.widget['exp'][index].observe(functools.partial(on_expression_toogle,
                                                            par=self, index=index),
                                          names='value')

def on_expression_toogle(c, par, index = 0):
    """
    Callback to expression checkbox
    """

    if c.new:
        par.widget['entry'][index] = par.widget['expression'][index]
        par.widget['expression'][index].layout.display = 'flex'
        par.widget['literal'][index].layout.display = 'none'
    else:
        par.widget['entry'][index] = par.widget['literal'][index]
        par.widget['expression'][index].layout.display = 'none'
        par.widget['literal'][index].layout.display = 'flex'

def on_add_clicked(button, par):
    """
    Callback to add button for multiple-value capable parameters
    """

    k = len(par.widget['entry'])
    par.widget['literal'].append(par.widget['entry_widget']())
    if par.has_placeholder():
        par.widget['literal'][k].placeholder = par.placeholder()

    par.widget['expression'].append(W.Text(placeholder="expression to evaluate"))
    par.widget['entry'].append(par.widget['literal'][k])

    par.widget['expression'][k].layout.display = 'none'

    par.widget['exp'].append(W.Checkbox(value=False, indent=False,
                                        layout = W.Layout(width="25px")))
    par.widget['exp'][k].observe(functools.partial(on_expression_toogle,
                                                   par=par, index = k),
                                 names='value')

    hbox = W.HBox([par.widget['literal'][k],
                   par.widget['expression'][k],
                   par.widget['exp'][k]], layout=W.Layout(width="100%"))
    children = par.widget['vbox'].children
    children = children + (hbox,)
    par.widget['vbox'].children = children

def on_del_clicked(button, par):
    """
    Callback to del button for multiple-value capable parameters
    """

    n = len(par.widget['entry'])
    if n == 1:
        return

    children = par.widget['vbox'].children
    par.widget['vbox'].children = children[:n-1]

    par.widget['literal'].pop(n-1)
    par.widget['expression'].pop(n-1)
    par.widget['exp'].pop(n-1)
    par.widget['entry'].pop(n-1)

def label_set_class(flag, label, css_class):

    if flag:
        label.add_class(css_class)



# This file is part of PyGeBR library.
#
# PyGeBR is a library to create, merge, edit, inspect and run
# processing flows. It was originally designed for seismic
# data processing, but it extends to other data as well.
#
# PyGeBR is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyGeBR is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyGeBR.  If not, see <https://www.gnu.org/licenses/>
#
# Copyright 2021-2022 Ricardo Biloti <biloti@unicamp.br>
#                     Eduardo Filpo <efilpo@gmail.com>
#

import ipywidgets as W
from .person import Person
from .param  import Param

#-------------------------------------------------------------------------------
class Prog:
    """Class used to represent a command line program

Attributes
----------

json: dict
    Dictionary representing a parameter

w: dict
    Dictionary holding an widget to edit programs's parameters

parlist: list of Param
    List of parameters which control the program behavior

authors: List of Person
    List of program's authors

title: str
    Program's title

description: str
    One-line description for the parameter purpose

executable: str
    Command to execute the program

stdin: boolean
    True if program reads from standar input

stdout: boolean
    True if program writes to standar output

stderr: boolean
    True if program writes to standar error

Below, in setter methods, providing "" as value has the effect to delete
the attribute. Note that a parameter has to have its type defined.
"""

    def __init__(self, json=None, title=None, description=None,
                 authors=None, url=None, executable=None,
                 stdin=True, stdout=True,stderr=True):
        """ Initiate a program from a JSON or empty otherwise.
        Some flow's properties may be further set.
        """

        self.parlist = []

        if json is None:
            self.json = {}

        else:
            if not isinstance(json, dict):
                raise TypeError("Argument should be a dictionary")

            self.json = json
            stdin = None
            stdout = None
            stderr = None

            # Construct the list of paramters from its json representation
            if 'parameters' in self.json:
                for parjs in self.json['parameters']:
                    self.parlist.append(Param(json = parjs))

        # Widgets for program
        self.w = {}

        if not title is None:
            self.title(title)
        if not description is None:
            self.description(description)
        if not authors is None:
            self.authors(authors)
        if not url is None:
            self.url(url)
        if not executable is None:
            self.executable(executable)
        if not stdin is None:
            self.stdin(stdin)
        if not stdout is None:
            self.stdout(stdout)
        if not stderr is None:
            self.stderr(stderr)

    def __str__(self):
        """
        Return a string to run the program from the
        command line, without resolving expressions though.

        To resolve expressions use self.eval().
        """

        cmdline = ""
        if self.has_executable():
            cmdline = self.executable()

        if self.has_parameters():
            for par in self.parameters():
                if not par.type() in ["section","comment"]:

                    if par.required() and not par.has_non_empty_value():
                        raise Exception("required parameter \'" + par.description() + "\' not set")

                    cmdline = " ".join([cmdline, str(par)]).strip()

        return cmdline

    def eval(self, params=None):
        """
        Return a string to run the program from the
        command line, resolving expressions.

        Parameter
        ---------

        param: dict (optional)
           Dictionary of variables to define local scope.
        """

        cmdline = ""
        if self.has_executable():
            cmdline = cmdline + self.executable()

        if self.has_parameters():
            for par in self.parameters():
                if not par.type() in ["section","comment"]:

                    if par.required() and not par.has_non_empty_value():
                        raise Exception("required parameter \'" + par.description() + "\' not set")

                    try:
                        parstr = par.eval(params)
                    except NameError as err:
                        raise NameError(str(err) + ", for program \'" + self.title() + "'")
                    except SyntaxError as err:
                        raise SyntaxError(str(err) + ", for program \'" + self.title() + "'" )
                    except ZeroDivisionError as err:
                        raise ZeroDivisionError(str(err) + ", for program \'" + self.title() + "'" )
                    except Exception as err:
                        raise Exception(str(err) + ", for program \'" + self.title() + "'" )

                    cmdline = " ".join([cmdline, parstr]).strip()

        return cmdline


    def has_title(self):
        """
        Return whether a program has title set
        """
        return 'title' in self.json

    def title(self, value=None):
        """
        Set or return program's title

        Parameter
        ---------

        value: (optinal)
            value to set program's title
        """
        if value is None:
            return self.json['title'] if self.has_title() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if (not value) and 'title' in self.json:
            del self.json['title']
        else:
            self.json['title'] = value

    def has_description(self):
        """
        Return whether a program has description set
        """
        return 'description' in self.json

    def description(self, value = None):
        """
        Set or return program's description

        Parameter
        ---------

        value: str (optinal)
            value to set program's description
        """
        if value is None:
            return self.json['description'] if self.has_description() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if (not value) and 'description' in self.json:
            del self.json['description']
        else:
            self.json['description'] = value

    def has_url(self):
        """
        Return whether a program has url set
        """
        return 'url' in self.json

    def url(self, value = None):
        """
        Set or return program's url

        Parameter
        ---------

        value: str (optinal)
            value to set program's url command
        """
        if value is None:
            return self.json['url'] if self.has_url() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if (not value):
            del self.json['url']
        else:
            self.json['url'] = value

    def has_executable(self):
        """
        Return whether a program has executable set
        """
        return 'executable' in self.json

    def executable(self, value = None):
        """
        Set or return program's executable

        Parameter
        ---------

        value: str (optinal)
            value to set program's executable command
        """
        if value is None:
            return self.json['executable'] if self.has_executable() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if (not value):
            del self.json['executable']
        else:
            self.json['executable'] = value

    def has_authors(self):
        """
        Return whether a program has authors set
        """
        return 'authors' in self.json

    def authors(self, value=None):
        """
        Set or return program's authors.

        Parameter
        ---------

        value: List of Person (optinal)
            value to set program's authors
        """
        if value is None:
            if not self.has_authors():
                return []
            authorslist = []
            for item in self.json['authors']:
                authorslist.append(Person(item))
            return authorslist

        if not isinstance(value, list) and not isinstance(value, Person):
            raise TypeError("Argument should be a list of persons")

        if isinstance(value, Person):
            self.json['authors'] = [value.json]
        else:
            if len(value) == 0:
                del self.json['authors']
            else:
                self.json['authors'] = []
                for person in value:
                    if not isinstance(person, Person):
                        raise Exception("Authors should be a list of persons")
                    self.json['authors'].append(person.json)

    def authors_count(self):
        """
        Return number of persons in authors' list
        """
        return len(self.authors())

    def stdin(self, value=None):
        """
        Query or set stdin acceptance for program

        Parameter
        ---------

        value: boolean (optional)
            True if program reads from the standard input
        """
        if value is None:
            if 'io' in self.json:
                return self.json['io']['stdin']

            return False

        if not isinstance(value, bool):
            raise TypeError("Argument should be boolean")

        if 'io' not in self.json:
            self.json['io'] = {}

        self.json['io']['stdin'] = value

    def stdout(self, value=None):
        """
        Query or set stdout write capability for program

        Parameter
        ---------

        value: boolean (optional)
            True if the program writes to standard ouput
        """
        if value is None:
            if 'io' in self.json:
                return self.json['io']['stdout']

            return False

        if not isinstance(value, bool):
            raise TypeError("Argument should be boolean")

        if 'io' not in self.json:
            self.json['io'] = {}

        self.json['io']['stdout'] = value

    def stderr(self, value=None):
        """
        Query or set stderr write capability for program

        Parameter
        ---------

        value: boolean (optional)
            True if the program writes to standard error
        """
        if value is None:
            if 'io' in self.json:
                return self.json['io']['stderr']

            return False

        if not isinstance(value, bool):
            raise TypeError("Argument should be boolean")

        if 'io' not in self.json:
            self.json['io'] = {}

        self.json['io']['stderr'] = value

    def has_parameters(self):
        """
        Return whether a programa has parameter list set
        """
        return 'parameters' in self.json

    def parameters(self):
        """
        Return the parameter list
        """
        return self.parlist if self.has_parameters() else []

    def parameters_count(self):
        """
        Return the number of parameters for the program
        """
        return len(self.parlist)

    def parameter(self, ipar):
        """
        Return the ipar-th program's parameter

        Parameter
        ---------

        ipar: int
            Parameter index ranging from 0 to self.parameters_count()-1
        """
        return self.parlist[ipar]

    def parameter_lookup(self, keyword):
        """
        Return the index for the first parameter
        such that the provided keyword matches keyword.

        Be aware that "scale" matches with "--xscale=". So, be
        as much specific as you can to avoid false positives.

        Parameter
        ---------

        keyword: str
            keyword to looking for
        """

        for k in range(self.parameters_count()):
            if keyword in self.parameter(k).keyword():
                return k

        return None

    def parameter_add(self, par):
        """
        Add a parameter to parameter's list for the program

        Parameter
        ---------

        par: Param
            Parameter to add
        """
        if not isinstance(par, Param):
            raise TypeError("Argument should be a Param")

        if len(self.parlist) == 0:
            self.json["parameters"] = []

        self.parlist.append(par)
        self.json["parameters"].append(par.json)

    def parameter_del(self, ipar):
        """
        Delete ipar-th parameter from parameter's list for the program

        Parameter
        ---------

        ipar: int
            Parameter index ranging from 0 to self.parameters_count()-1
        """
        self.json['parameters'].pop(ipar)
        self.parlist.pop(ipar)
        if len(self.parlist) == 0:
            del self.json['parameters']

    def widget(self, status):
        """
        Return a widget to represent the program

        Parameter
        ---------

        status: boolean
            True to set program as active
        """
        if self.w == {}:
            self.widget_construct(status)

        return self.w['main']

    def widget_construct(self,status):
        """
        Construct the widget

        Parameter
        ---------

        status: boolean
            True to set program as active
        """

        stack = []

        # Program's description
        linecontent = []
        if self.has_description():
            desc = W.Label(value=self.description())
            desc.add_class("prog_desc")
            linecontent.append(desc)

        # Link to authors' homepage
        if self.has_url():
            links = W.HTML(value='<a href="' + self.url() +
                           '" target="_blank">' +
                           '<i class="fa fa-external-link" aria-hidden="true"></i></a>')
            linecontent.append(links)

        if linecontent:
            stack.append(W.HBox(linecontent))

        # Active checkbox
        self.w['status'] = W.Checkbox(description="Enable this program in the flow",
                                      indent=False,
                                      layout = W.Layout(width="100%"),
                                      value = status)

        stack.append(W.HBox([self.w['status']]))

        inside = False
        first = True
        accordion = None
        has_required = False
        for par in self.parlist:
            has_required = has_required or par.required()
            if par.type() == 'section':
                accordion = par.w_line()
                stack.append(accordion)
                vbox = accordion.children[0]
                inside = True
                first = True
            else:
                if inside:
                    if first:
                        vbox.children = (par.w_line(),)
                        first = False
                    else:
                        vbox.children = vbox.children + (par.w_line(),)
                else:
                    stack.append(par.w_line())

        if has_required:
            footnote = W.Label(value="* required parameter")
            footnote.add_class("footnote")
            stack.append(footnote)
        self.w['main'] = W.VBox(stack)

    def widget_value_set(self):
        """
        Fill widget values from program/parameter's properties
        """

        for par in self.parameters():
            par.w_set()

    def widget_value_get(self):
        """
        Fill program/parameter's properties from widget's values
        """

        for par in self.parameters():
            par.w_get()



# This file is part of PyGeBR library.
#
# PyGeBR is a library to create, merge, edit, inspect and run
# processing flows. It was originally designed for seismic
# data processing, but it extends to other data as well.
#
# PyGeBR is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyGeBR is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyGeBR.  If not, see <https://www.gnu.org/licenses/>
#
# Copyright 2021-2022 Ricardo Biloti <biloti@unicamp.br>
#                     Eduardo Filpo <efilpo@gmail.com>
#

import json as JSON
import subprocess
import functools
from time import sleep
import os

from colorama import Fore, Style
from jsonschema import Draft202012Validator
import ipywidgets as W
from IPython.core.display import display, HTML

from .person  import Person
from .program import Prog
from .setup   import Setup

#--------------------------------------------------------------------------------
def _dump_parameter(par, isfull, setonly, resolve, pardict, parcount):
    """ Private function to dump a parameter
    """

    ptype = par.type()
    if ptype == 'section':
        print("      " + Fore.GREEN + par.formated(isfull, resolve, pardict)+":" + Style.RESET_ALL)
        return

    if ptype == 'comment':
        print("      " + Fore.YELLOW + par.formated(isfull, resolve, pardict) + Style.RESET_ALL)
        if isfull:
            print("")
        return

    if setonly and not par.has_non_empty_value():
        return

    output = par.formated(isfull, resolve, pardict)
    first = True

    for line in output:
        if first:
            print("      [%2d] "%parcount + line)
            first = False
        else:
            print("           " + line)

    if isfull:
        print("")

#-------------------------------------------------------------------------------
class Flow:
    """Class to represent a processing flow.

A flow is a representation of a processing,
in which data flows through a pipe of programs.

Internally, a flow is a python dictionary, save/exported
as a JSON.

A flow can be initialized from another flow, eventually read from
a file.

Optionaly, a dictionary can be provided to evaluate
parameters defined by expressions.

Attributes
----------

json: dict
    Dictionary representing the flow

widget: dict
    Dictionary holding an widget to setup, inspect and run the flow

filename: str
    Filename from which the flow is loaded or saved

par: dict
    Dictionary of variables to define local scope to evaluate parameters' expressions

authors: List of Person
    List of program's authors

title: str
    Program's title

description: str
    One-line description for the parameter purpose

tags: List of str
    List of tags for catalog purpose

stdin: str
    Filename feed to standard input

stdout: str
    Filename to write the standard output

stderr: str
    Filename to write the standard error

"""

    schemafn = os.path.join(os.path.dirname(__file__),
                            "schema",
                            "full-flow.schema")
    validator = None

    def __init__(self, flow=None, filename=None, par=None,
                 title=None, description=None,
                 authors=None, tags=None,
                 stdin=None, stdout=None, stderr=None):
        """
        Attributes
        ----------

        flow: dict (optional)
            dictionary representing the flow

        filename: str (optional)
            Filename from which the flow is loaded

        par: dict (optional)
            Dictionary of variables to define local scope to evaluate parameters' expressions

        authors: List of Person (optional)
            List of program's authors

        title: str (optional)
            Program's title

        description: str (optional)
            One-line description for the parameter purpose

        tags: List of str (optional)
            List of tags for catalog purpose

        stdin: str (optional)
            Filename feed to standard input

        stdout: str (optional)
            Filename to write the standard output

        stderr: str (optional)
            Filename to write the standard error
        """

        if not par is None:
            self.pardict = par
        else:
            self.pardict = {}

        self.json = {}
        self.proglist = []
        self.authorlist = []
        self.w = {}

        if flow is not None:
            if not isinstance(flow, Flow):
                raise TypeError("Argument should be Flow")
            self.json = JSON.loads(JSON.dumps(flow.json))
        elif not filename is None:
            with open(filename) as fp:
                text = fp.read()

            self.json = JSON.loads(text)
            self.filename(filename)

        # Recreate Programs and Parameters
        if self.json != {} and 'programs' in self.json:
            for item in self.json['programs']:
                self.proglist.append(Prog(json = item['program']))

        # Recreate Authors
        if self.json != {} and 'authors' in self.json:
            for item in self.json['authors']:
                self.authorlist.append(Person(json = item))

        if not title is None:
            self.title(title)
        if not description is None:
            self.description(description)
        if not authors is None:
            self.authors(authors)
        if not tags is None:
            self.tags(tags)
        if not stdin is None:
            self.stdin(stdin)
        if not stdout is None:
            self.stdout(stdout)
        if not stderr is None:
            self.stderr(stderr)

    def save(self, filename=None):
        """
        Save the flow as a JSON to disk

        Parameter
        ---------

        filename: str (optional)
           Filename to save the flow in. If not provided
           the flow will be saved in self.filename().
        """

        if filename is None:
            if self.has_filename():
                fn = self.filename()
            else:
                raise Exception("Filename nor set neither provided")
        else:
            fn = filename

        with open(fn, "w") as fp:
            fp.write(JSON.dumps(self.json))

        self.filename(fn)

    def copy(self):
        """
        Copy the flow preserving reference to dictionary of parameters
        """
        flow = Flow(flow=self, par=self.pardict)
        flow.filename("")
        return flow

    def has_par(self):
        """
        Return whether the flow has dictionary of parameters
        """
        if self.pardict:
            return True
        else:
            return False

    def par(self, value=None):
        """
        Set or return flow's dictionary of parameters

        Parameter
        ---------

        value: dict (optional)
            Dictionary containing variables to define local scope
            when evaluating parameter's expressions (leave it blank for querying)
        """
        if value is None:
            return self.pardict

        if not isinstance(value, dict):
            raise TypeError("Argument should be a dictionary")

        self.pardict = value

    def has_title(self):
        """
        Return whether the flow has title set
        """
        return 'title' in self.json

    def title(self, value = None):
        """
        Set or return flow's title

        Parameter
        ---------

        value: str (optional)
            Short title for the flow (leave it blank for querying)
        """
        if value is None:
            return self.json['title'] if self.has_title() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if (not value) and 'title' in self.json:
            del self.json['title']
        else:
            self.json['title'] = value

    def has_description(self):
        """
        Return whether the flow has description set
        """
        return 'description' in self.json

    def description(self, value = None):
        """
        Set or return flow's description

        Parameter
        ---------

        value: str (optional)
            One-sentence description for the flow (leave it blank for querying)
        """
        if value is None:
            return self.json['description'] if self.has_description() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if (not value) and 'description' in self.json:
            del self.json['description']
        else:
            self.json['description'] = value

    def has_tags(self):
        """
        Return whether the flow has tags set
        """
        return 'tags' in self.json

    def tags(self, value = None):
        """
        Set or return flow's tags

        Parameter
        ---------

        value: list of str (optional)
            List of tags (leave it blank for querying)
        """
        if value is None:
            return self.json['tags'] if self.has_tags() else []

        if not isinstance(value, list):
            raise TypeError("Argument should be a list of strings")

        if (not value) and 'tags' in self.json:
            del self.json['tags']
        else:
            self.json['tags'] = value

    def accept_stdin(self):
        """
        Return whether flow reads from stdin
        """
        nprogs = self.programs_count()

        k = 0
        while k < nprogs:
            if self.program_status(k):
                return self.program(k).stdin()
            k = k + 1
        return False

    def accept_stdout(self):
        """
        Return whether flow writes to stdout
        """

        nprogs = self.programs_count()

        k = nprogs -1
        while k >=0:
            if self.program_status(k):
                return self.program(k).stdout()
            k = k - 1
        return False

    def accept_stderr(self):
        """
        Return whether flow writes to stderr
        """

        nprogs = self.programs_count()

        k = nprogs -1
        while k >=0:
            if self.program_status(k):
                return self.program(k).stderr()
            k = k - 1
        return False

    def has_authors(self):
        """
        Return whether the flow has authors set
        """
        return 'authors' in self.json

    def authors(self, value=None):
        """
        Set or return flow's authors.

        Parameter
        ---------

        value: (optional)
            A list of Person (leave it blank for querying)
        """
        if value is None:
            return self.authorlist if self.has_authors() else []

        if not isinstance(value, list) and not isinstance(value, Person):
            raise TypeError("Argument should be a list of persons")

        if isinstance(value, Person):
            self.json['authors'] = [value.json]
            self.authorlist = [value]
        else:
            if (not value) and 'authors' in self.json:
                del self.json['authors']
            else:
                self.json['authors'] = []
                for person in value:
                    if not isinstance(person, Person):
                        raise Exception("Authors should be a list of persons")
                    self.json['authors'].append(person.json)
                self.authorlist = value

    def authors_count(self):
        """ Return the number of persons in credit's list
        """
        return len(self.authors())

    def has_filename(self):
        """
        Return whether the flow has filename set
        """
        return 'filename' in self.json

    def filename(self, value = None):
        """
        Set or return flow's filename

        Parameter
        ---------

        value: str (optional)
            Filename to save/load the flow (leave it blank for querying)
        """
        if value is None:
            return self.json['filename'] if self.has_filename() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if (not value) and 'filename' in self.json:
            del self.json['filename']
        else:
            self.json['filename'] = value

        return 'filename' in self.json

    def has_stdin(self):
        """ Return whether the flow has stdin set
        """
        if 'io' in self.json:
            return 'stdin' in self.json['io']

        return False

    def stdin(self, value=None):
        """Set or return flow's stdin
        """
        if value is None:
            return self.json['io']['stdin'] if self.has_stdin() else ""

        if not isinstance(value, str):
            raise TypeError("Argument should be a string")

        if not value:
            if 'io' in self.json:
                if 'stdin' in self.json['io']:
                    del self.json['io']['stdin']
                if not self.json['io']:
                    del self.json['io']
        else:
            if 'io' not in self.json:
                self.json['io']={}

            self.json['io']['stdin'] = value

    def has_stdout(self):
        """ Return whether the flow has stdout set
        """
        if 'io' in self.json:
            return 'stdout' in self.json['io']

        return False

    def stdout(self, value=None):
        """
        Set or return flow's stdout.

        When set, provide a tuple with filename and mode ('a' or 'w').
        For example,
        flow.stdout(['/tmp/data.rsf','w'])
        """
        if value is None:
            return self.json['io']['stdout'] if self.has_stdout() else []

        if (not isinstance(value, list)) and (not value == ""):
            raise TypeError("Argument should be a list, like ['filename','w']")

        if not value:
            if 'io' in self.json:
                if 'stdout' in self.json['io']:
                    del self.json['io']['stdout']
                if not self.json['io']:
                    del self.json['io']
        else:
            if 'io' not in self.json:
                self.json['io']={}

            self.json['io']['stdout'] = value

    def has_stderr(self):
        """
        Return whether the flow has stderr set
        """
        if 'io' in self.json:
            return 'stderr' in self.json['io']

        return False

    def stderr(self, value=None):
        """
        Set or return flow's stderr.

        When set, provide a list with filename and mode ('a' or 'w').
        For example,
        flow.stderr(['/tmp/log.txt','a'])
        """
        if value is None:
            return self.json['io']['stderr'] if self.has_stderr() else []

        if (not isinstance(value, list)) and (not value == ""):
            raise TypeError("Argument should be a list, like ['filename','w']")

        if not value:
            if 'io' in self.json:
                if 'stderr' in self.json['io']:
                    del self.json['io']['stderr']
                if not self.json['io']:
                    del self.json['io']
        else:
            if 'io' not in self.json:
                self.json['io']={}

            self.json['io']['stderr'] = value

    def has_programs(self):
        """
        Return whether the flow has programs
        """
        return 'programs' in self.json

    def programs(self):
        """
        Return the list of programs for the flow
        """
        return self.proglist

    def programs_count(self):
        """
        Return the number of programs in the flow
        """
        return len(self.proglist)

    def program_status(self, iprog, status=None):
        """
        Set or return the iprog-th program's status

        Parameter
        ----------

        iprog: int
            Index of the program

        status: boolean
            Indicates with the program should be
            added as active (True) or inactive (False)
        """
        if (iprog < 0 or iprog >= self.programs_count()):
            raise ValueError("Program index out of range")

        if status is None:
            return self.json['programs'][iprog]['enabled']

        if not isinstance(status, bool):
            raise TypeError("Argument should be a boolean")

        self.json['programs'][iprog]['enabled']  = status

    def program(self, iprog):
        """
        Return the iprog-th program in flow

        Parameter
        ---------

        iprog: int
            Index of the program
        """
        if (iprog < 0 or iprog >= self.programs_count()):
            raise ValueError("Program index out of range")

        return self.proglist[iprog]

    def program_add(self, prog, status=True):
        """
        Add a program to the flow

        Parameters
        ----------

        prog: Prog
            Program to be added

        status: boolean
            Indicates with the program should be
            added as active (True) or inactive (False)
        """

        if not isinstance(prog, Prog):
            raise TypeError("Argument should be a Prog")

        if len(self.proglist) == 0:
            self.json['programs'] = []

        self.proglist.append(prog)
        self.json['programs'].append({"enabled": status, "program": prog.json})

    def program_del(self, iprog):
        """ Delete i-th program in flow

        Parameter
        ---------

        iprog: int
            Index of the program to be deleted
        """

        if (iprog < 0 or iprog >= self.programs_count()):
            raise ValueError("Program index out of range")

        self.json['programs'].pop(iprog)
        self.proglist.pop(iprog)
        if len(self.proglist) == 0:
            del self.json['programs']

    def program_parameter_values(self, iprog, ipar, value = None):
        """ Set or return the value of ipar-th parameter
        of iprog-th program of flow

        Parameter
        ---------

        iprog: int
            Index of the program to be deleted

        ipar: int
            Index of the parameter to set or query

        value: (optional)
            Value to set (leave empty to query)
        """
        if (iprog < 0 or iprog >= self.programs_count()):
            raise ValueError("Program index out of range")

        if (ipar < 0 or ipar >= self.program(iprog).parameters_count()):
            raise ValueError("Parameter index out of range")

        param = self.proglist[iprog].parameter(ipar)

        if value is None:
            return param.values()

        param.values(value)

    def validate(self,verbose=False):
        """
        Validate the flow against schema

        Parameter
        ---------

        verbose: bool (optional)
            If True, besides returning whether the flow is
            valid, returns the list of errors.
        """
        if self.validator is None:
            with open(Flow.schemafn,'r') as fp:
                Flow.validator = Draft202012Validator(JSON.loads(fp.read()))

        valid = Flow.validator.is_valid(self.json)
        if (not verbose):
            return valid

        errorlist = []
        for error in sorted(Flow.validator.iter_errors(self.json),key=str):
            errorlist.append(error.message)

        return valid,errorlist

    def dump(self,verbose=False,setonly=False,resolve=False):
        """
        Dump the flow's content

        Parameter
        ---------

        verbose: boolean (optional)
            Be more verbose than usual

        setonly: boolean (optional)
            Show only parameters with value set

        resolve: boolean (optional)
            Resolve parameters' values defined by expression
        """

        # Title -- Description
        headtext = self.title()
        if self.has_description():
            headtext = headtext + " — " + self.description()

        print(headtext)

        # Authors
        if self.has_authors():
            for person in self.authors():
                print(str(person)+"\n")

        # Tags
        if self.has_tags():
            tags = self.tags()
            ntags = len(tags)
            for k in range(ntags-1):
                print(Fore.CYAN + tags[k] + Style.RESET_ALL, end=" ∙ ")

            print(Fore.CYAN + tags[ntags-1] + Style.RESET_ALL+"\n")

        # IO: stdin, stdout and stderr
        iotext = ""
        if self.has_stdin():
            iotext = iotext + "<" + self.stdin() + "  "

        if self.has_stdout():
            filename, mode = self.stdout()
            if mode == "w":
                iotext = iotext + ">" + filename + "  "
            else:
                iotext = iotext + ">>" + filename + "  "

        if self.has_stderr():
            filename, mode = self.stderr()
            if mode == "w":
                iotext = iotext + "2>" + filename
            else:
                iotext = iotext + "2>>" + filename

        if iotext:
            print(iotext)

        # Programs
        print("Flow with " + str(self.programs_count()) + " program(s):")

        for k in range(0, self.programs_count()):
            prog = self.program(k)

            headtext = "\n[%2i]: "%k + Fore.GREEN + prog.title() + Style.RESET_ALL
            if prog.has_description():
                headtext = headtext + " — " + prog.description()
            headtext = headtext + " (" + prog.executable() + ")"

            print(headtext)

            if self.program_status(k):
                if prog.has_parameters():
                    print(Fore.GREEN + "      Parameters:" + Style.RESET_ALL)
                    parcount = 0
                    for parcount in range(prog.parameters_count()):
                        _dump_parameter(prog.parameter(parcount),verbose,setonly,resolve,self.pardict, parcount)
                else:
                    print("      No parameters")
            else:
                print("      Disabled")

    def __str__(self):
        """
        Return the command line to run the flow,
        without resolving expressions though.

        To resolve parameters use self.eval().
        """
        return self.__cmdline(False)

    def eval(self):
        """
        Return the command line to run the flow,
        resolving expressions.
        """
        return self.__cmdline(True)


    def __cmdline(self, resolve):
        """
        Return the command line to run the flow.

        Parameter
        ---------

        resolve: boolean
            Resolve or not parameter's valued defined by expressions
        """
        nprogs = self.programs_count()
        if nprogs == 0:
            return ""

        cmdline = ""
        k = 0
        first = True
        for k in range(nprogs):
            if self.program_status(k):
                prog = self.program(k)
                if first:
                    if prog.stdin() and not self.has_stdin():
                        raise Exception("First program reads from stdin")

                    if prog.stdin():
                        cmdline = "<" + self.stdin() + " "

                    if resolve:
                        cmdline = cmdline + prog.eval(self.par())
                    else:
                        cmdline = cmdline + str(prog)

                    first = False
                else:
                    if ((self.program(k).stdin() and not self.program(k-1).stdout()) or
                        (not self.program(k).stdin() and self.program(k-1).stdout())):
                        raise Exception("Broken flow between %s and %s"%
                                        (self.program(k-1).executable(),
                                         self.program(k).executable()))

                    if (self.program(k).stdin() and self.program(k-1).stdout()):
                        if resolve:
                            cmdline = cmdline + " | " + prog.eval(self.par())
                        else:
                            cmdline = cmdline + " | " + str(prog)

                    if (not self.program(k).stdin() and not self.program(k-1).stdout()):
                        if resolve:
                            try:
                                progstr = prog.eval(self.par())
                            except:
                                raise

                            cmdline = cmdline + " && " + progstr
                        else:
                            cmdline = cmdline + " && " + str(prog)

            k = k + 1

        if first:
            return ""

        if self.has_stdout():
            nprog = nprogs - 1
            while nprog >= 0:
                if self.program_status(nprog):
                    break

                nprog = nprog - 1

            if nprog >= 0:
                prog = self.program(nprog)
                if prog.stdout():
                    stdout = self.stdout()
                    if stdout[1] == "w":
                        cmdline = cmdline + " >"
                    elif stdout[1] == "a":
                        cmdline = cmdline + " >>"
                    else:
                        raise Exception("Unknown stdout mode")

                    cmdline = cmdline + stdout[0]

        if self.has_stderr():
            nprog = nprogs - 1
            while nprog >=0 :
                if self.program_status(nprog):
                    break

                nprog = nprog - 1

            if nprog >= 0:
                prog = self.program(nprog)
                if prog.stderr():
                    stderr = self.stderr()
                    if stderr[1] == "w":
                        cmdline = cmdline + " 2>"
                    elif stderr[1] == "a":
                        cmdline = cmdline + " 2>>"
                    else:
                        raise Exception("Unknown stderr mode")

                    cmdline = cmdline + stderr[0]

        return cmdline

    def merge(self, flow):
        """
        Merge flow

        Parameter
        ---------

        flow: Flow
            A flow to merge into self
        """
        self.json['programs'] = self.json['programs'] + flow.json['programs']
        for prog in flow.proglist:
            self.proglist.append(prog)

        # Destroy the widget to rebuild it when necessary
        self.w = {}

    def run(self):
        """
        Run the flow

        """
        accept_stdout = self.accept_stdout()
        accept_stderr = self.accept_stderr()

        has_stdout = self.has_stdout()
        has_stderr = self.has_stderr()

        if accept_stdout:
            if has_stdout:
                path = os.path.dirname(self.stdout()[0])
                if not os.path.exists(path):
                    os.mkdir(path)
                stdout = open(self.stdout()[0], self.stdout()[1])
            else:
                stdout = subprocess.PIPE
        else:
            stdout = None

        if accept_stderr:
            if has_stderr:
                path = os.path.dirname(self.stderr()[0])
                if not os.path.exists(path):
                    os.mkdir(path)
                stderr = open(self.stderr()[0], self.stderr()[1])
            elif not has_stdout and accept_stdout:
                stderr = subprocess.STDOUT
            else:
                stderr = subprocess.PIPE
        else:
            stderr = None

        output = ""
        cmdstr = None
        try:
            cmdstr = self.eval()
        except:
            raise

        if cmdstr:
            run = subprocess.run([cmdstr], shell=True, text=True,
                                 stdout=stdout, stderr=stderr)

            if accept_stdout:
                if has_stdout:
                    stdout.close()
                else:
                    output = output + run.stdout

            if accept_stderr:
                if has_stderr:
                    stderr.close()
                else:
                    if (run.stderr):
                        output = output + run.stderr

        return output

    def _widget_io_update(self):
        """
        Enable or disable standard in/out/err
        and filled it in widget
        """

        if self.accept_stdin():
            self.w['stdin'].disabled = False
            if self.has_stdin():
                self.w['stdin'].value = self.stdin()
            else:
                self.w['stdin'].value = ''
        else:
            self.w['stdin'].disabled = True

        if self.accept_stdout():
            self.w['stdout'].disabled = False
            if self.has_stdout():
                self.w['stdout'].value = self.stdout()[0]
            else:
                self.w['stdout'].value = ''
        else:
            self.w['stdout'].disabled = True

        if self.accept_stderr():
            self.w['stderr'].disabled = False
            if self.has_stderr():
                self.w['stderr'].value = self.stderr()[0]
            else:
                self.w['stderr'].value = ''
        else:
            self.w['stderr'].disabled = True


    def widget_populate(self):
        """
        Populate flow's widget
        """

        self.w['title'].value = self.title()
        self.w['description'].value = self.description()

        self._widget_io_update()

        self.w['filename'].value = self.filename()
        k = 0
        for prog in self.programs():
            prog.w['status'].value = self.program_status(k)
            prog.widget_value_set()
            k = k + 1

    def widget_construct(self):
        """
        Construct the widget to inspect, edit and run the flow
        """

        display(HTML("""
        <style>
            .flow_title {font-weight: bold;}
            .flow_desc  {font-style: italic;}
            .prog_desc  {font-style: italic; padding-right: 10px;}
            .section {font-weight: bold;}
            .required {font-weight: bold;}
            .required::after {content: "*"}
            .footnote {font-style: italic;}
            .flowoutput {font-family: monospace;}
            .commentpar {line-height: 14pt;
                         padding-bottom: 5px;
                         padding-top: 5px;
                         color: #006680;}
        </style>"""))

        if self.w != {}:
            self.widget_populate()
            return

        #----------------------------------------------------------------------------------------
        # Side bar
        label_title = W.Label(value = self.title())
        label_title.add_class("flow_title")
        self.w['title'] = W.Text(placeholder="Flow's title")
        link_title = W.link((label_title,'value'),(self.w['title'],'value'))

        label_description = W.Label(value = self.description())
        label_description.add_class("flow_desc")
        self.w['description'] = W.Text(placeholder="Flow's description")
        link_description = W.link((label_description,'value'),(self.w['description'],'value'))

        label_files = W.HTML(value='Files <i class="fa fa-files-o" aria-hidden="true"></i>',
                             layout=W.Layout(margin="10px 0px 0px 0px"))
        label_files.add_class("section")

        self.w['stdin'] = W.Text(placeholder="Input file")
        self.w['stdout'] = W.Text(placeholder="Output file")
        self.w['stderr'] = W.Text(placeholder="Error log")

        bt_info = W.Button(description="Info", button_style = "info", icon="info")
        bt_info.on_click(functools.partial(on_info_clicked, flow=self))

        bt_commit = W.Button(description="Commit", button_style="info", icon="download")
        bt_commit.on_click(functools.partial(on_commit_clicked, flow=self))

        self.w['filename'] = W.Text(placeholder="File to save this flow")
        bt_save = W.Button(description="Save",button_style="info", icon="floppy-o")
        bt_save.on_click(functools.partial(on_save_clicked, flow=self))

        bt_run = W.Button(description="Run",button_style="danger", icon="check")
        bt_run.on_click(functools.partial(on_run_clicked, flow=self))

        #----------------------------------------------------------------------------------------
        # Programs
        self.w['programs'] = W.Tab(layout=W.Layout(width="67%",padding="0px 2px 0px 5px"))

        #----------------------------------------------------------------------------------------
        # Info bar
        label_titdesc = W.Label(value = "Title and description")
        label_titdesc.add_class("section")

        label_authors = W.HTML(value='Authors <i class="fa fa-id-card-o" aria-hidden="true"></i>',
                               layout=W.Layout(margin="10px 0px 0px 0px"))
        label_authors.add_class("section")
        children = (label_titdesc, W.HBox([self.w['title'], self.w['description']]), label_authors,)

        for person in self.authors():
            children = children + (person.w_line(),)

        self.w['info'] = W.VBox(children, layout=W.Layout(width="67%", padding="0px 2px 0px 5px"))
        self.w['info'].layout.display = 'none'

        #----------------------------------------------------------------------------------------
        # Main window
        self.w['main'] = W.HBox([W.VBox([label_title,label_description, label_files,
                                         self.w['stdin'],
                                         self.w['stdout'],
                                         self.w['stderr'],
                                         self.w['filename'],
                                         W.HBox([bt_info,bt_commit],layout=W.Layout(margin="10px 0px 0px 0px")),
                                         W.HBox([bt_save,bt_run])],
                                         layout=W.Layout(widht="27%",padding="0px 2px 0px 5px")),
                         self.w['info'],
                         self.w['programs']],
                         layout=W.Layout(widht="100%",border="1px solid"))
        self.w['output'] = W.Textarea(layout=W.Layout(width="100%", height="98%"))
        self.w['output'].add_class("flowoutput")

        if len(self.proglist) > 0:
            children = (self.program(0).widget(self.program_status(0)),)
            for k in range(1,self.programs_count()):
                children = children + (self.program(k).widget(self.program_status(k)),)
            children = children + (self.w['output'],)
            self.w['programs'].children = children

        k = 0
        for prog in self.programs():
            self.w['programs'].set_title(k, prog.title())
            prog.w['status'].observe(functools.partial(on_program_status_changed, flow=self, k=k), names='value')
            k = k + 1
        self.w['programs'].set_title(k, "Flow's output")

        # After constructing the widget, populate it
        self.widget_populate()

    def w_get(self):
        """
        Set flow's properties from widget value
        """

        self.title(self.w['title'].value)
        self.description(self.w['description'].value)
        self.filename(self.w['filename'].value)

        if not self.w['stdin'].disabled:
            self.stdin(self.w['stdin'].value)

        if not self.w['stdout'].disabled:
            if self.w['stdout'].value:
                self.stdout([self.w['stdout'].value,"w"])
            else:
                self.stdout("")

        if not self.w['stderr'].disabled:
            if self.w['stderr'].value:
                self.stderr([self.w['stderr'].value,"w"])
            else:
                self.stderr("")

        for person in self.authors():
            person.w_get()

        for prog in self.programs():
            prog.widget_value_get()

    def w_set(self):
        """
        Set widget value from flow's properties
        """

        self.w['title'].value = self.title()
        self.w['description'].value = self.description()
        self.w['filename'].value = self.filename()

        self._widget_io_update()

        for person in self.authors():
            person.w_set()

        for prog in self.programs():
            prog.widget_value_set()

    def ui(self):
        """
        Construct and display flow's widget
        """
        self.widget_construct()
        display(self.w['main'])

def on_commit_clicked(button, flow):
    """
    Callback to commit button
    """

    flow.w_get()

def on_info_clicked(button, flow):
    """
    Callback to info button
    """

    if flow.w['info'].layout.display == 'none':
        flow.w['info'].layout.display = 'flex'
        flow.w['programs'].layout.display = 'none'

        button.description = "Programs"
        button.button_style = "warning"
        button.icon = "cubes"

    else:
        flow.w['info'].layout.display = 'none'
        flow.w['programs'].layout.display = 'flex'

        button.description = "Info"
        button.button_style = "info"
        button.icon = "info"

def on_run_clicked(button, flow):
    """
    Callback to run button
    """

    button.button_style = "warning"
    button.icon = "gear"
    flow.w_get()
    try:
        output = flow.run()
    except Exception as err:
        flow.w['output'].value = "ERROR: "+str(err)
        flow.w['programs'].selected_index = flow.programs_count()
        button.button_style = "danger"
        button.icon = "check"
        return

    flow.w['output'].value = output
    flow.w['programs'].selected_index = flow.programs_count()
    button.button_style = "danger"
    button.icon = "check"

def on_save_clicked(button, flow):
    """
    Callback to save button
    """

    if flow.w['filename'].value:
        flow.w_get()
        try:
            flow.save(flow.w['filename'].value)
        except Exception as err:
            flow.w['output'].value = "Error: " + str(err)
            flow.w['programs'].selected_index = flow.programs_count()
            return

        button.button_style = "success"
        sleep(0.4)
        button.button_style = "info"

    else:
        button.button_style = "danger"
        sleep(0.4)
        button.button_style = "info"

def on_program_status_changed(c, flow, k):
    """
    Callback to program status checkbox
    """

    flow.program_status(k, c.new)
    flow._widget_io_update()
